import os
import sys
import glob
import time
import shutil

from log import logger
from load import S3Loader


POLL_INTERVAL = 60


class LogListener(object):
    
    def __init__(self, log_dir, bucket, region, remove=False, poll=POLL_INTERVAL):
        """
        Listen for changes to files in log_dir

        When TwitterStream rotates a log the event handler will be called
        The LogEventS3Handler will load the log file into a s3 bucket on AWS
        """
        self.log_dir = log_dir
        self.bucket = bucket
        self.s3 = S3Loader(bucket, region)
        self.poll = poll
        self.remove = remove
        try:
            self.s3.valid()
        except EnvironmentError as e:
            logger.error(e)
            sys.exit(1)

    def start(self):
        if not os.path.isdir(self.log_dir):
            msg = "Directory < {0} > does not exist".format(self.log_dir)
            logger.error(msg)
            sys.exit(1)

        pattern = 'twitter-stream.log.*'
        while True:
            for log in glob.glob(os.path.join(self.log_dir, pattern)):
                self.load(log)
            time.sleep(self.poll)

    def load(self, log):
        if not self.s3:
            return
        msg = "Loading {0} to S3 bucket {1}".format(log, self.bucket)
        logger.info(msg)
        self.s3.store(log)
        logger.info("Success loading {0}".format(log))
        self.archive(log)

    def archive(self, log):
        if not self.remove:
            filepath = os.path.dirname(log)
            filename = os.path.basename(log)
            archive = os.path.join(filepath, 'loaded.{0}'.format(filename))
            logger.info("Archiving {0} to {1}".format(log, archive))
            shutil.copyfile(log, archive)
        logger.info("Removing log file {0}".format(log))
        os.remove(log)



import os
import sys
import json
import argparse

root = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.append(root)

from log import logger
from stream import TwitterStream
from listen import LogListener
from steps import GeoFilterStep, ExtractStep, LogStep


# Bounding Box: [Lon,Lat SW corner, Lon,Lat NE corner]
PORTLAND = "-123.784828,44.683842,-121.651703,46.188969"
CONTINENTAL_US = "-125.0011,24.9493,-66.9326,49.5904"

# default parameters from environment variables or set to defaults
DEFAULT_LOG_DIR = "/tmp/geotweet"
DEFAULT_LOG_INTERVAL = 5
DEFAULT_LOCATION = CONTINENTAL_US


class Geotweet(object):
    
    handler = None
    starting_step = None

    def __init__(self, args):
        self.log_dir = args.log_dir if args.log_dir else DEFAULT_LOG_DIR
        try:
            # stream arguments
            self.log_interval = args.log_interval \
                if args.log_interval else DEFAULT_LOG_INTERVAL
            self.bbox = self.build_bbox(args.bbox if args.bbox else CONTINENTAL_US)
        except AttributeError:
            pass
        try:
            # listen arguments
            self.bucket = args.bucket
            self.region = args.region
        except AttributeError:
            pass
    
    def build_bbox(self, bbox):
        if not bbox:
            return None
        # build bounding box from bbox argument
        coords = [ coord.strip() for coord in bbox.split(',') ]
        if len(coords) != 4:
            error = "--bbox must be in the form 'sw-x,sw-y,ne-x,ne-y'"
            raise ValueError(error.format(bbox))
        return [coords[0] + "," + coords[1], coords[2] + "," + coords[3]]

    def add_step(self, step):
        if not self.starting_step:
            self.starting_step = step
            return
        tmp = self.starting_step
        while tmp.get_next():
            tmp = tmp.get_next()
        else:
            tmp.set_next(step)
   
    def stream(self):
        try:
            os.mkdir(self.log_dir)
        except OSError:
            msg = "Output log directory < {0} > already exists".format(self.log_dir)
            logger.debug(msg)
        log = os.path.join(self.log_dir, 'twitter-stream.log')
        logger.info("Starting Twitter Streaming API")
        logger.info("Streaming to output log: {0}".format(log))
        logger.info("Log Interval (min): {0}".format(self.log_interval))
        logger.info("Bounding Box: {0}".format(self.bbox))
        # initialize processing step chain
        self.add_step(GeoFilterStep())
        self.add_step(ExtractStep())
        self.add_step(LogStep(logfile=log, log_interval=self.log_interval))
        # start Twitter Streaming API (does not return)
        TwitterStream().start(process_step=self.starting_step, locations=self.bbox)

    def listen(self):
        msg = "Start listening for events in directory: {0}"
        logger.info(msg.format(self.log_dir))
        logger.info("AWS Bucket: {0}".format(self.bucket))
        logger.info("AWS Region: {0}".format(self.region))
        # start listener (does not return)
        LogListener(self.log_dir, self.bucket, self.region).start()



import json

from log import logger, get_rotating_logger


class ProcessStep(object):
    """ Base Class used to build data processing chain """ 
    next_step = None
   
    def get_next(self):
        return self.next_step
    
    def set_next(self, next_step):
        self.next_step = next_step

    def next(self, data):
        if self.next_step:
            return self.next_step.process(data)
        return data

    def process(self, step_input):
        """ Process input and and return output """ 
        return self.next(step_input)


class GeoFilterStep(ProcessStep):
    """
    Process output from Twitter Streaming API
    
    For each record output from the API will be called as argument to process.
    That function will validate and convert tweet to desired format.
    
    """
    def _validate(self, key, record):
        if key in record and record[key]:
            return True
        return False

    def validate_geotweet(self, record):
        """ check that stream record is actual tweet with coordinates """
        if record and  self._validate('user', record) \
                and self._validate('coordinates', record):
            return True
        return False

    def process(self, tweet):
        """ Passes on tweet if missing 'geo' or 'user' property """
        if self.validate_geotweet(tweet):
            return self.next(tweet)
        return None


class ExtractStep(ProcessStep):
    """ Extract interesting fields from Tweet """
    
    def process(self, tweet):
        if not tweet:
            return None
        user = tweet['user']
        data = dict(
            user_id=user['id'],
            name=user['name'],
            screen_name=user['screen_name'],
            description=user['description'],
            location=user['location'],
            friends_count=user['friends_count'],
            followers_count=user['followers_count'],
            tweet_id=tweet['id_str'],
            source=tweet['source'],
            created_at=tweet['created_at'],
            timestamp=tweet['timestamp_ms'],
            lonlat=tweet['coordinates']['coordinates']
        )
        return self.next(data)


class LogStep(ProcessStep):
    """ Log tweet to rotating log """

    def __init__(self, logfile, log_interval=60, when="M"):
        if logfile:
            self.rotating_logger = get_rotating_logger(logfile, log_interval)

    def process(self, tweet):
        if tweet:
            self.rotating_logger.info(json.dumps(tweet))
            return self.next(tweet) 



import os

import boto3


class S3Loader(object):
    
    envvars = ['AWS_ACCESS_KEY_ID', 'AWS_SECRET_ACCESS_KEY', 'AWS_DEFAULT_REGION']

    def __init__(self, bucket=None, region=None):
        self.bucket = bucket
        if region:
            os.environ['AWS_DEFAULT_REGION'] = region
    
    def valid(self):
        if not self.bucket:
            error = "Error: AWS Bucket not set"
            raise ValueError(error)
        for envvar in self.envvars:
            if not os.getenv(envvar):
                error = "Error: Environment variable {0} not set".format(envvar)
                raise EnvironmentError(error)

    def store(self, filepath):
        if not self.bucket:
            return False
        filename = filepath.rsplit('/', 1)[-1]
        s3 = boto3.resource('s3')
        s3.Object(self.bucket, filename).put(Body=open(filepath, 'rb'))
